import os
from typing import List, Optional

import typer
import yaml

from savvihub.cli.constants import PROJECT_TYPE_VERSION_CONTROL, PROJECT_TYPE_CLI_DRIVEN
from savvihub.cli.exceptions import ExitException
from savvihub.cli.typer import Context
from savvihub.cli.utils import inquire_list, inquire_text


def experiment_config_file_callback(ctx: Context, filepath: str) -> Optional[str]:
    if not filepath:
        return

    if not os.path.isfile(filepath):
        raise ExitException(f'File does not exist: {filepath}')

    try:
        yaml.safe_load(filepath)
    except yaml.YAMLError:
        raise ExitException(f'Invalid YAML: {filepath}')

    with open(filepath, 'r') as stream:
        configs = yaml.load(stream, Loader=yaml.FullLoader)

    ctx.spec['experiment'] = configs['spec']['experiment']


def message_callback(ctx: Context, message: str) -> str:
    if message:
        ctx.store['message'] = message
        return ctx.store['message']

    if ctx.spec['experiment'].get('message'):
        ctx.store['message'] = ctx.spec['experiment'].get('message')
        return ctx.store['message']

    inquirer_message = ""
    if ctx.project.type == PROJECT_TYPE_VERSION_CONTROL:
        inquirer_message = "Message(set to commit message if you passed empty)"
    elif ctx.project.type == PROJECT_TYPE_CLI_DRIVEN:
        inquirer_message = "Message"

    message = inquire_text(inquirer_message)
    ctx.store['message'] = message
    return ctx.store['message']


def start_command_callback(ctx: Context, start_command: str) -> str:
    if start_command:
        ctx.store['start_command'] = start_command
        return ctx.store['start_command']

    if ctx.spec['experiment'].get('start_command'):
        ctx.store['start_command'] = ctx.spec['experiment'].get('start_command')
        return ctx.store['start_command']

    start_command = inquire_text("Start command", "python main.py")
    ctx.store['start_command'] = start_command
    return ctx.store['start_command']


def cluster_name_callback(ctx: Context, cluster_name: str) -> str:
    assert ctx.params['organization_name']

    clusters = ctx.authenticated_client.cluster_list(ctx.params['organization_name']).clusters
    clusters = {c.name: c for c in clusters if c.status == 'connected'}

    cluster_name = cluster_name.strip()
    if not cluster_name:
        cluster_name = ctx.spec['experiment'].get('cluster')
    
    if cluster_name:
        if cluster_name in clusters:
            ctx.store['cluster'] = clusters[cluster_name]
        else:
            raise ExitException(f'Cluster not found: {cluster_name}')

    elif len(clusters) == 1:
        selected_cluster = list(clusters.values())[0]
        ctx.store['cluster'] = selected_cluster
        typer.echo(
            f'The cluster is automatically set to '
            f'`{selected_cluster.name}{" (SavviHub)" if selected_cluster.is_savvihub_managed else f" (Custom)"}`.'
        )

    else:
        selected_cluster = inquire_list(
            "Please choose a cluster",
            [(f'{x.name}{" (SavviHub)" if x.is_savvihub_managed else f" ({x.name})"}', x) for x in clusters.values()]
        )
        ctx.store['cluster'] = selected_cluster

    return ctx.store['cluster'].name


def image_url_callback(ctx: Context, image_url: str) -> str:
    assert ctx.params['organization_name'] and ctx.params['processor_type']

    image_url = image_url.strip()
    if image_url:
        ctx.store['image_url'] = image_url
        return image_url

    image_url = ctx.spec['experiment'].get('image_url')

    images = ctx.authenticated_client.kernel_image_list(ctx.params['organization_name']).results
    images = {i.image_url: i for i in images if i.processor_type == ctx.params['processor_type']}

    if not image_url:
        selected_image = inquire_list(
            "Please choose a kernel image",
            [(f'{x.image_url} ({x.name})', x.image_url) for x in images.values()],
        )
        ctx.store['image_url'] = selected_image
    
    return ctx.store['image_url']


def env_vars_callback(ctx: Context, env_vars: List[str]) -> List[str]:
    if ctx.spec['experiment'].get('env_vars'):
        if not isinstance(ctx.spec['experiment'].get('env_vars'), List):
            raise ExitException(f'Invalid environment variables in YAML')
        env_vars += tuple(ctx.spec['experiment']['env_vars'])

    ctx.store['env_vars'] = []
    for env_var in env_vars:
        try:
            env_key, env_value = env_var.split("=", 1)
            ctx.store['env_vars'].append({
                'key': env_key,
                'value': env_value,
            })
        except ValueError:
            raise ExitException(f'Cannot parse environment variable: {env_var}')
    
    return ctx.store['env_vars']



import numbers
from typing import Mapping, List

import six

from openapi_client import ResponseExperimentInfo
from savvihub import history
from savvihub.exceptions import ArgumentException
from savvihub.objects import Image


class Experiment:
    def __init__(self, experiment: ResponseExperimentInfo, client):
        self.id = experiment.id
        self.client = client
        self._history = None
        self._plot_volume_id = experiment.experiment_plot_volume

    def get_plot_volume_id(self):
        return self._plot_volume_id

    def refine(self, row, step):
        # Row validation
        if not isinstance(row, Mapping):
            raise ArgumentException(".log() takes a dictionary as a parameter")

        if any(not isinstance(key, six.string_types) for key in row.keys()):
            raise ArgumentException("The key of dictionary in .log() parameter must be str")

        for k in row.keys():
            if not k:
                raise ArgumentException("Logging empty key is not supported")

        # Step validation
        if step is not None:
            if not isinstance(step, numbers.Number):
                raise ArgumentException(f"Step must be a number, not {type(step)}")

            if not isinstance(type(step), int):
                step = int(round(step))

        return row, step

    @classmethod
    def from_given(cls, experiment_id, client):
        return cls(client.experiment_id_read(experiment_id), client)

    @property
    def history(self):
        if not self._history:
            self._history = history.History(self)
        return self._history

    def log(self, row, *, step=None):
        row, step = self.refine(row, step)
        for val in row.values():
            if isinstance(val, List) and len(val) > 0 and all(isinstance(i, Image) for i in val):
                self.history.update_images(self.client, row)
                break
            else:
                self.history.update_metrics(self.client, row, step)
                break



import os
from typing import List

from openapi_client import (
    ProtoVolumeMountRequest,
    ProtoVolumeMountRequestSourceDataset,
    ProtoVolumeMountRequestSourceDatasetVersion,
)
from savvihub.api.exceptions import NotFoundAPIException
from savvihub.cli.exceptions import ExitException
from savvihub.cli.typer import Context
from savvihub.cli.utils import parse_dataset
from savvihub.cli.constants import DATASET_VERSION_HASH_LATEST


def dataset_mount_callback(ctx: Context, dataset_mounts: List[str]) -> List[str]:
    client = ctx.authenticated_client
    organization_name = ctx.params['organization_name']
    ctx.store['dataset_mounts'] = []
    for dataset_mount in dataset_mounts:
        splitted = dataset_mount.split(':')
        if len(splitted) != 2:
            raise ExitException(f'Invalid dataset mount format: {dataset_mount}. '
                                f'You should specify both mount path and dataset name.\n'
                                f'ex) /input/dataset1:mnist@3d1e0f91c')

        mount_path, dataset_full_name = splitted
        mount_path = os.path.join(mount_path, '')  # Forcibly add slash to mount path
        organization_name_override, dataset_name, dataset_version_hash = parse_dataset(dataset_full_name)
        if organization_name_override:
            organization_name = organization_name_override

        dataset = client.dataset_read(organization_name, dataset_name)
        if dataset.is_version_enabled:
            if dataset_version_hash != DATASET_VERSION_HASH_LATEST:
                # Validate dataset version
                try:
                    dataset_version = client.dataset_version_read(dataset.id, dataset_version_hash)
                    dataset_version_hash = dataset_version.version_hash  # Replace with full version hash
                except NotFoundAPIException:
                    raise ExitException(f'Invalid dataset version: {dataset_full_name}\n'
                                        f'Please check your dataset version in `{organization_name}`.')
            ctx.store['dataset_mounts'].append(ProtoVolumeMountRequest(
                mount_type='dataset-version',
                mount_path=mount_path,
                dataset_version=ProtoVolumeMountRequestSourceDatasetVersion(
                    dataset_id=dataset.id,
                    dataset_name=dataset_name,
                    dataset_version_hash=dataset_version_hash,
                ),
            ))
        else:
            ctx.store['dataset_mounts'].append(ProtoVolumeMountRequest(
                mount_type='dataset',
                mount_path=mount_path,
                dataset=ProtoVolumeMountRequestSourceDataset(
                    dataset_id=dataset.id,
                    dataset_name=dataset_name,
                ),
            ))

    return dataset_mounts



import os
import tempfile
from typing import Optional

import inquirer
import typer
import requests

from savvihub.api.file_object import DownloadableFileObject
from savvihub.api.uploader import Uploader
from savvihub.cli.constants import PROJECT_TYPE_CLI_DRIVEN
from savvihub.cli.exceptions import ExitException
from savvihub.cli.typer import Context


def git_ref_callback(ctx: Context, git_ref: str) -> Optional[str]:
    if ctx.project.type == PROJECT_TYPE_CLI_DRIVEN:
        return None

    if git_ref:
        git_ref = git_ref.strip()
        if ctx.git_repo and not ctx.git_repo.check_revision_in_remote(git_ref):
            raise ExitException(f'Git commit {git_ref} does not exist in a remote repository.')
        return git_ref

    if ctx.git_repo is None:
        raise ExitException('Run it inside a git-initialized directory or set --git-ref option.')

    return ctx.git_repo.commit_ref


def git_branch_callback(ctx: Context, git_branch: str) -> Optional[str]:
    if ctx.project.type == PROJECT_TYPE_CLI_DRIVEN:
        return None

    if git_branch:
        return git_branch

    if ctx.git_repo is None:
        raise ExitException('Run it inside a git-initialized directory or set --git-branch option.')

    return ctx.git_repo.branch


def git_diff_callback(ctx: Context, git_diff_url: str) -> Optional[str]:
    if ctx.project.type == PROJECT_TYPE_CLI_DRIVEN:
        return None

    assert ctx.params['git_ref']
    if ctx.params['ignore_git_diff']:
        return None

    diff_file = None
    if git_diff_url:
        if git_diff_url.startswith('https://') or git_diff_url.startswith('http://'):
            diff_file = tempfile.NamedTemporaryFile(suffix='.patch')
            d = DownloadableFileObject(git_diff_url, os.path.dirname(diff_file.name), os.path.basename(diff_file.name))
            session = requests.Session()
            session.headers = ctx.authenticated_client.auth_header
            d.download(session=session)
            diff_file.seek(0)
    else:
        if ctx.git_repo is None:
            return None

        typer.echo(f'Run experiment with revision {ctx.git_repo.commit_ref[:7]} ({ctx.git_repo.branch})')
        typer.echo(f'Commit: {ctx.git_repo.get_commit_message(ctx.git_repo.commit_ref)}')
        if not ctx.git_repo.is_head:
            typer.echo('Your current revision does not exist in remote repository. '
                       'SavviHub will use latest remote branch revision hash and uncommitted diff.')
        typer.echo('')

        has_diff, diff_status = ctx.git_repo.get_current_diff_status(ctx.git_repo.commit_ref)
        if has_diff:
            typer.echo('Diff to be uploaded: ')

            uncommitted_files = diff_status.get('uncommitted')
            untracked_files = diff_status.get('untracked')

            if uncommitted_files:
                typer.echo('  Changes not committed')
                typer.echo('\n'.join([f'    {x}' for x in uncommitted_files]))
                typer.echo('')
            if untracked_files:
                typer.echo(f'  Untracked files:')
                typer.echo('\n'.join([f'    {x}' for x in untracked_files]))
                typer.echo('')

            answer = inquirer.prompt([inquirer.List(
                'question',
                message='Run experiment with diff?',
                choices=[
                    ('[1] Run experiment with uncommitted and untracked changes.', 1),
                    ('[2] Run experiment with uncommitted changes.', 2),
                    ('[3] Run experiment without any changes.', 3),
                    ('[4] Abort.', 4),
                ],
            )], raise_keyboard_interrupt=True)['question']

            if answer == 1:
                diff_file = ctx.git_repo.get_current_diff_file(ctx.git_repo.commit_ref, with_untracked=True)
            elif answer == 2:
                diff_file = ctx.git_repo.get_current_diff_file(ctx.git_repo.commit_ref, with_untracked=False)
            elif answer == 3:
                pass
            else:
                raise ExitException('Aborted.')

    if diff_file:
        typer.echo('Generating diff patch file...')
        uploaded = Uploader.upload(
            ctx.authenticated_client,
            local_path=diff_file.name,
            volume_id=ctx.project.volume_id,
            remote_path=os.path.basename(diff_file.name),
            progressable=typer.progressbar,
        )
        diff_file_path = uploaded.path
        diff_file.close()

        ctx.store['git_diff_uploaded_path'] = diff_file_path

    return git_diff_url



from datetime import datetime
from typing import List

import typer
from terminaltables import AsciiTable

from openapi_client import (
    ModelKernelResourceSpecField,
    ProtoVolumeMountRequest,
    ProtoVolumeMountRequestSourceCLIDrivenProject,
    ProtoVolumeMountRequestSourceProject,
    ProtoVolumeMountRequestSourceVersionControlProject,
    ProtoVolumeMountRequests,
)
from savvihub.cli.commands.volume import volume_file_list, volume_file_copy
from savvihub.cli.constants import WEB_HOST, PROJECT_TYPE_VERSION_CONTROL, PROJECT_TYPE_CLI_DRIVEN
from savvihub.cli.exceptions import ExitException
from savvihub.cli.formatter import TreeFormatter
from savvihub.cli.inputs.experiment import (
    cluster_name_callback,
    env_vars_callback,
    image_url_callback,
    message_callback,
    start_command_callback,
)
from savvihub.cli.inputs.experiment import experiment_config_file_callback
from savvihub.cli.inputs.git import git_branch_callback, git_ref_callback, git_diff_callback
from savvihub.cli.inputs.organization import organization_name_option
from savvihub.cli.inputs.project import project_name_option, local_project_callback
from savvihub.cli.inputs.resource import (
    cpu_limit_callback,
    gpu_limit_callback,
    gpu_type_callback,
    memory_limit_callback,
    processor_type_callback,
    resource_name_callback,
)
from savvihub.cli.inputs.volume import dataset_mount_callback
from savvihub.cli.typer import Typer, Context
from savvihub.common.utils import (
    parse_time_to_ago,
    short_string,
)

experiment_output_app = Typer()
experiment_app = Typer()
experiment_app.add_typer(experiment_output_app, name='output')


@experiment_app.callback()
def main():
    """
    Run the machine learning experiment
    """
    return


def parse_resource_spec(resource_spec):
    if not resource_spec.name:
        resource_desc = "[Custom]"
        if resource_spec.processor_type == "gpu":
            resource_desc += "g" + str(int(resource_spec.gpu_limit)) + "."
        resource_desc += "c" + str(int(resource_spec.cpu_limit)) + "."
        resource_desc += "r" + resource_spec.memory_limit[:-2]
        return resource_desc
    return resource_spec.name


@experiment_app.command(user_required=True)
def list(
    ctx: Context,
    organization_name: str = organization_name_option,
    project_name: str = project_name_option,
):
    """
    Display a list of experiments
    """
    experiments = ctx.authenticated_client.experiment_list(organization_name, project_name).results
    if not experiments:
        raise ExitException(f'There is no experiment in `{project_name}` project.')

    table = AsciiTable([
        ['NUMBER', 'NAME', 'STATUS', 'CREATED', 'RESOURCE', 'MESSAGE'],
        *[[e.number, e.name, e.status, parse_time_to_ago(e.created_dt),
           parse_resource_spec(e.kernel_resource_spec),
           f'"{short_string(e.message, 25)}"']
          for e in experiments],
    ])
    table.inner_column_border = False
    table.inner_heading_row_border = False
    table.inner_footing_row_border = False
    table.outer_border = False

    typer.echo(table.table)


@experiment_app.command(user_required=True)
def describe(
    ctx: Context,
    organization_name: str = organization_name_option,
    project_name: str = project_name_option,
    experiment_number_or_name: str = typer.Argument(..., help="The unique experiment number or name"),
):
    """
    Describe the experiment in details
    """
    experiment = ctx.authenticated_client.experiment_read(organization_name, project_name, experiment_number_or_name)
    timezone = datetime.now().astimezone().tzinfo

    root = TreeFormatter()
    root.add_child(f'Number: {experiment.number}',
                   f'Name: {experiment.name}',
                   f'Created: {experiment.created_dt.astimezone(timezone)}',
                   f'Updated: {experiment.updated_dt.astimezone(timezone)}',
                   f'Git Commit: ({experiment.git_ref[:7]}) {experiment.message}')

    if experiment.source_code_link:
        root.add_child(f'Source code link: {experiment.source_code_link[0].url}')

    if experiment.git_diff_file:
        git_diff_file_desc = TreeFormatter('Git Diff File:')
        git_diff_file_desc.add_child(f'URL: {experiment.git_diff_file.download_url["url"]}')
        root.add_child(git_diff_file_desc)

    root.add_child(f'Status: {experiment.status}',
                   f'Tensorboard: {experiment.tensorboard or "N/A"}')

    # Kernel Image Description
    kernel_image_desc = TreeFormatter('Kernel Image:')
    kernel_image_desc.add_child(f'Name: {experiment.kernel_image.name}',
                                f'URL: {experiment.kernel_image.image_url}',
                                f'Language: {experiment.kernel_image.language or "N/A"}')
    root.add_child(kernel_image_desc)

    # Resource Spec Description
    resource_spec_desc = TreeFormatter('Resource Spec:')
    resource_spec_desc.add_child(f'Name: {experiment.kernel_resource_spec.name}',
                                 f'CPU Type: {experiment.kernel_resource_spec.name}',
                                 f'CPU Limit: {experiment.kernel_resource_spec.cpu_limit}',
                                 f'Memory Limit: {experiment.kernel_resource_spec.memory_limit}',
                                 f'GPU Type: {experiment.kernel_resource_spec.gpu_type}',
                                 f'GPU Limit: {experiment.kernel_resource_spec.gpu_limit}')
    root.add_child(resource_spec_desc)

    # Datasets Description
    datasets_desc = TreeFormatter('Datasets:')
    found = False
    for volume_mount in experiment.volume_mounts.mounts:
        if volume_mount.source_type == 'dataset':
            dataset_desc = TreeFormatter(volume_mount.dataset.dataset.name)
            dataset_desc.add_child(f'Mount Path: {volume_mount.path}')
            datasets_desc.add_child(dataset_desc)
            found = True
    if not found:
        datasets_desc.add_child('Dataset Not Found')
    root.add_child(datasets_desc)

    # Histories Description
    histories_desc = TreeFormatter('Histories:')
    if not experiment.histories:
        histories_desc.add_child('History Not Found')
    else:
        for history in experiment.histories:
            history_desc = TreeFormatter(history.status)
            history_desc.add_child(f'Started: {datetime.fromtimestamp(history.started_timestamp, tz=timezone).strftime("%Y-%m-%d %H:%M:%S.%f%z")}',
                                   f'Ended: {datetime.fromtimestamp(history.ended_timestamp, tz=timezone).strftime("%Y-%m-%d %H:%M:%S.%f%z") if history.ended_timestamp else "N/A"}')
            histories_desc.add_child(history_desc)
    root.add_child(histories_desc)

    # Plots Description
    plots_desc = TreeFormatter('Plots:')
    full_plots_info_desc = TreeFormatter('Full plots at:')
    full_plots_info_desc.add_child(f'{WEB_HOST}/{organization_name}/{project_name}/experiments/{experiment.number}/plots')
    plots_desc.add_child(full_plots_info_desc)
    root.add_child(plots_desc)
    root.add_child(f'Start Command: {experiment.start_command}')

    typer.echo(root.format())


@experiment_app.command(user_required=True)
def logs(
    ctx: Context,
    organization_name: str = organization_name_option,
    project_name: str = project_name_option,
    experiment_number_or_name: str = typer.Argument(..., help="The unique experiment number or name"),
    tail: int = typer.Option(200, "--tail"),
    detail: bool = typer.Option(False, "--detail", hidden=True),
    all: bool = typer.Option(False, "--all", hidden=True),
):
    """
    Display the last fifty lines of the experiment logs
    """
    client = ctx.authenticated_client

    kwargs = {}
    if not all:
        kwargs = {'limit': tail}
    if detail:
        kwargs = {'with_event_log': 'true'}

    experiment_logs = client.experiment_logs(
        organization_name, project_name, experiment_number_or_name, **kwargs).logs

    all_logs = []
    for _, logs in experiment_logs.items():
        all_logs.extend(logs)

    timezone = datetime.now().astimezone().tzinfo
    log_str = ''
    for log in sorted(all_logs, key=lambda x: x.timestamp):
        log_str += f'[{datetime.fromtimestamp(log.timestamp, tz=timezone).strftime("%H:%M:%S.%f")}] {log.message}\n'

    typer.echo(log_str)
    typer.echo(
        f'Full logs at:\n'
        f'    {WEB_HOST}/{organization_name}/{project_name}/experiments/{experiment_number_or_name}/logs\n'
    )


@experiment_app.command(user_required=True)
def run(
    ctx: Context,
    organization_name: str = organization_name_option,
    project_name: str = project_name_option,
    file: str = typer.Option(None, "--file", "-f", callback=experiment_config_file_callback, help="Experiment config file"),
    message: str = typer.Option("", "--message", "-m", callback=message_callback, help="Experiment message"),
    start_command: str = typer.Option("", "--start-command", callback=start_command_callback,
                                      help="Start command"),
    cluster_name: str = typer.Option("", "--cluster", "-c", callback=cluster_name_callback,
                                     help="Cluster name"),
    resource_name: str = typer.Option(None, "--resource", "-r", callback=resource_name_callback,
                                      help="Resource name (for savvihub-managed cluster)"),
    processor_type: str = typer.Option(None, "--processor-type", callback=processor_type_callback,
                                       help="cpu or gpu (for custom cluster)"),
    cpu_limit: float = typer.Option(0, "--cpu-limit", callback=cpu_limit_callback,
                                    help="Number of vCPUs (for custom cluster)"),
    memory_limit: str = typer.Option(None, "--memory-limit", callback=memory_limit_callback,
                                     help="Memory capacity (ex: 4Gi, 500Mi)"),
    gpu_type: str = typer.Option(None, "--gpu-type", callback=gpu_type_callback,
                                 help="GPU product name such as Tesla-K80 (for custom cluster)"),
    gpu_limit: int = typer.Option(0, "--gpu-limit", callback=gpu_limit_callback,
                                  help="Number of GPU cores (for custom cluster)"),
    image_url: str = typer.Option("", "--image", "-i", callback=image_url_callback,
                                  help="Kernel docker image URL"),
    dataset_mounts: List[str] = typer.Option([], "--dataset", "-d", callback=dataset_mount_callback,
                                             help="Dataset mounted path"),
    env_vars: List[str] = typer.Option([], "-e", callback=env_vars_callback, help="Environment variables"),
    ignore_git_diff: bool = typer.Option(False, "--ignore-git-diff", help="Ignore git diff flag"),
    git_branch: str = typer.Option(None, "--git-branch", callback=git_branch_callback, help="Git branch name"),
    git_ref: str = typer.Option(None, "--git-ref", callback=git_ref_callback, help="Git commit SHA"),
    git_diff_arg: str = typer.Option(None, "--git-diff", callback=git_diff_callback, help="Git diff file URL"),
    local_project: str = typer.Option(None, "--local-project", callback=local_project_callback,
                                      help="Local project file URL"),
    output_dir: str = typer.Option("/output/", "--output-dir",
                                   help="A directory to which the experiment result output files to be stored."),
    working_dir: str = typer.Option(None, "--working-dir", help="If not present, use `/work/{project_name}`."),
    root_volume_size: str = typer.Option(None, "--root-volume-size", help="Root volume size"),
):
    """
    Run an experiment in SavviHub
    """
    resource_spec = resource_spec_id = None
    if ctx.store['cluster'].is_savvihub_managed:
        resource_spec_id = ctx.store['resource_spec_id']
    else:
        resource_spec = ModelKernelResourceSpecField(
            processor_type=processor_type,
            cpu_type='Any',
            cpu_limit=cpu_limit,
            memory_limit=memory_limit,
            gpu_type=gpu_type,
            gpu_limit=gpu_limit,
        )

    project_volume_mount = ProtoVolumeMountRequestSourceProject(
        project_id=ctx.project.id,
    )
    project_mount_path = "/work/"
    if ctx.project.type == PROJECT_TYPE_VERSION_CONTROL:
        project_volume_mount.project_type = PROJECT_TYPE_VERSION_CONTROL
        project_volume_mount.version_control_project = ProtoVolumeMountRequestSourceVersionControlProject(
            project_branch=git_branch,
            project_git_ref=git_ref,
            project_git_diff=ctx.store.get('git_diff_uploaded_path')
        )
        project_mount_path += f"{project_name}"
    elif ctx.project.type == PROJECT_TYPE_CLI_DRIVEN:
        project_volume_mount.project_type = PROJECT_TYPE_CLI_DRIVEN
        project_volume_mount.cli_driven_project = ProtoVolumeMountRequestSourceCLIDrivenProject(
            local_project=ctx.store.get('local_project_uploaded_path')
        )

    volume_mounts_requests = ctx.store['dataset_mounts'] + [
        ProtoVolumeMountRequest(
            mount_type='output',
            mount_path=output_dir,
        ),
        ProtoVolumeMountRequest(
            mount_type='project',
            mount_path=project_mount_path,
            project=project_volume_mount,
        ),
    ]
    if project_mount_path != '/work/':
        volume_mounts_requests.append(ProtoVolumeMountRequest(
            mount_type='empty-dir',
            mount_path='/work/'
        ))

    experiment = ctx.authenticated_client.experiment_create(
        organization=organization_name,
        project=project_name,
        message=message,
        cluster_name=ctx.store['cluster'].name,
        image_url=image_url,
        resource_spec_id=resource_spec_id,
        resource_spec=resource_spec,
        start_command=start_command,
        env_vars=ctx.store['env_vars'],
        volume_mounts=ProtoVolumeMountRequests(
            root_volume_size=root_volume_size,
            working_dir=working_dir,
            requests=volume_mounts_requests,
        ),
    )

    typer.echo(
        f'Experiment {experiment.number} is running. Check the experiment status at below link\n'
        f'{WEB_HOST}/{organization_name}/{project_name}/experiments/{experiment.number}'
    )


@experiment_output_app.callback()
def output_main():
    """
    Manage experiment output files
    """


@experiment_output_app.command(user_required=True)
def ls(
    ctx: Context,
    organization_name: str = organization_name_option,
    project_name: str = project_name_option,
    experiment_number_or_name: str = typer.Argument(..., help="The unique experiment number or name"),
    path: str = typer.Argument(None, help='Output file path'),
    recursive: bool = typer.Option(False, '-r', '--recursive', help='recursive flag'),
    directory: bool = typer.Option(False, '-d', '--directory',
                                   help='list the directory itself, not its contents'),
):
    """
    List the output files of the experiment
    """
    experiment = ctx.authenticated_client.experiment_read(organization_name, project_name, experiment_number_or_name)
    if len(experiment.volume_mounts.mounts) > 0:
        for volume_mount in experiment.volume_mounts.mounts:
            if volume_mount.source_type == 'output':
                volume_file_list(ctx, volume_mount.volume.volume.id, 'latest', path or '', recursive, directory)
    else:
        raise ExitException('No output volume mounted.')


@experiment_output_app.command(user_required=True)
def download(
    ctx: Context,
    organization_name: str = organization_name_option,
    project_name: str = project_name_option,
    experiment_number_or_name: str = typer.Argument(..., help="The unique experiment number or name"),
    dest_path: str = typer.Argument(None, help='The files will be downloaded to ./output if omitted.'),
):
    """
    Download experiment output files
    """
    experiment = ctx.authenticated_client.experiment_read(organization_name, project_name, experiment_number_or_name)
    if len(experiment.volume_mounts.mounts) > 0:
        for volume_mount in experiment.volume_mounts.mounts:
            if volume_mount.source_type == 'output':
                volume_file_copy(
                    ctx, volume_mount.volume.volume.id,
                    '', None, dest_path or './output',
                    recursive=True, watch=False)
                break
    else:
        raise ExitException('No output volume mounted.')



import ast
import os
from enum import Enum
from typing import List, Optional

import yaml

from savvihub.cli.exceptions import ExitException
from savvihub.cli.typer import Context
from savvihub.cli.utils import (
    inquire_float,
    inquire_int,
    inquire_list,
    inquire_text,
    validate_enum,
    validate_float,
    validate_int,
    validate_text,
)


class SweepAlgorithmEnum(str, Enum):
    grid = "grid"
    random = "random"
    bayesian = "bayesian"


class SweepObjectiveTypeEnum(str, Enum):
    maximize = "maximize"
    minimize = "minimize"


class SweepSearchSpaceParameterTypeEnum(str, Enum):
    int = "int"
    float = "float"
    double = "double"
    categorical = "categorical"


def sweep_config_file_callback(ctx: Context, filepath: str) -> Optional[str]:
    if not filepath:
        return

    if not os.path.isfile(filepath):
        raise ExitException(f'File does not exist: {filepath}')

    try:
        yaml.safe_load(filepath)
    except yaml.YAMLError:
        raise ExitException(f'Invalid YAML: {filepath}')

    with open(filepath, 'r') as stream:
        configs = yaml.load(stream, Loader=yaml.FullLoader)

    ctx.spec['experiment'] = configs['spec']['experiment']
    ctx.spec['sweep'] = configs['spec']['sweep']
    return


def algorithm_callback(ctx: Context, algorithm: str) -> str:
    valid_algorithms = [a.value for a in SweepAlgorithmEnum]

    if algorithm:
        ctx.store['algorithm'] = algorithm.lower()
        validate_enum(
            ctx.store['algorithm'],
            SweepAlgorithmEnum, 
            f'Invalid sweep algorithm: {algorithm}. (choose from {", ".join(valid_algorithms)})'
        )
        return ctx.store['algorithm']

    ctx.store['algorithm'] = inquire_list(
        "Please select an algorithm",
        [(x.capitalize(), x) for x in valid_algorithms],
    )
    return ctx.store['algorithm']


def objective_type_callback(ctx: Context, objective_type: str) -> str:
    valid_objective_types = [t.value for t in SweepObjectiveTypeEnum]

    if objective_type:
        ctx.store['objective_type'] = objective_type.lower()
        validate_enum(
            ctx.store['objective_type'],
            SweepObjectiveTypeEnum, 
            f'Invalid objective type: {objective_type}. (choose from {", ".join(valid_objective_types)})'
        )
        return ctx.store['objective_type']

    ctx.store['objective_type'] = inquire_list(
        "Please choose an objective type",
        [(x.capitalize(), x) for x in valid_objective_types],
    )
    return ctx.store['objective_type']


def objective_goal_callback(ctx: Context, objective_goal: float) -> float:
    ctx.store['objective_goal'] = objective_goal

    if ctx.store['objective_goal'] is None:
        ctx.store['objective_goal'] = inquire_float("Objective goal (between 0 and 1)")

    validate_float(
        ctx.store['objective_goal'],
        f'Invalid objective goal: {objective_goal}. Please enter a decimal between 0 and 1.',
        gte=0, lte=1,
    )
    return ctx.store['objective_goal']


def objective_metric_callback(ctx: Context, objective_metric: str) -> str:
    ctx.store['objective_metric'] = objective_metric
    if not ctx.store['objective_metric']:
        ctx.store['objective_metric'] = inquire_text("Objective metric")

    validate_text(ctx.store['objective_metric'], "Objective metric cannot be blank.")
    return ctx.store['objective_metric']


def _parse_parameter(unparsed_parameter, valid_parameter_types):
    parameter = ast.literal_eval(unparsed_parameter)

    # Parameter name
    parameter["name"] = parameter.get("name")
    validate_text(parameter["name"], "Parameter name cannot be blank.")
    
    # Parameter type
    parameter["type"] = parameter.get("type")
    validate_enum(
        parameter["type"],
        SweepSearchSpaceParameterTypeEnum, 
        f'Invalid parameter type {parameter["type"]}. (choose from {", ".join(valid_parameter_types)}) '
    )

    # Parameter values
    parameter_range = parameter.get("range", {})
    parameter["range"]["list"] = parameter_range.get("list")
    parameter["range"]["min"] = parameter_range.get("min")
    parameter["range"]["max"] = parameter_range.get("max")

    # Parameter values: list
    if parameter["range"]["list"] or parameter["type"] == SweepSearchSpaceParameterTypeEnum.categorical:
        if not isinstance(parameter["range"]["list"], list) or len(parameter["range"]["list"]) == 0:
            raise ExitException("Invalid parameter range values.")

        if parameter["type"] == SweepSearchSpaceParameterTypeEnum.int:
            for v in parameter["range"]["list"]:
                validate_int(v, "Invalid parameter range values.")
        elif parameter["type"] in [SweepSearchSpaceParameterTypeEnum.float, SweepSearchSpaceParameterTypeEnum.double]:
            for v in parameter["range"]["list"]:
                validate_float(v, "Invalid parameter range values.")

    # Parameter values: min/max
    elif parameter["range"]["min"] and parameter["range"]["max"]:
        validate_int(parameter["range"]["min"], "Invalid parameter range values.")
        validate_int(parameter["range"]["max"], "Invalid parameter range values.", gt=parameter["range"]["min"])

    else:
        raise ExitException("Invalid parameter range type (must be min/max or list).")
    
    return parameter


def search_space_parameters_callback(ctx: Context, unparsed_parameters: List[str]) -> Optional[List[str]]:
    search_space_parameters = []
    valid_parameter_types = [t.value for t in SweepSearchSpaceParameterTypeEnum]

    # Via options
    if unparsed_parameters:
        for unparsed_parameter in unparsed_parameters:
            parameter = _parse_parameter(unparsed_parameter, valid_parameter_types)
            search_space_parameters.append(parameter)

        ctx.store["search_space"] = search_space_parameters
        return

    # Via inquirer
    while True:
        parameter = {}
        parameter["range"] = {}

        count = len(search_space_parameters) + 1
        message_prefix = f'Search space parameter #{count}'

        # Parameter name
        parameter["name"] = inquire_text(f'{message_prefix} - name')
        validate_text(parameter["name"], "Parameter name cannot be blank.")

        # Parameter type
        parameter["type"] = inquire_list(f'{message_prefix} - type', valid_parameter_types)

        # Parameter range type (min_max / list)
        if parameter["type"] == SweepSearchSpaceParameterTypeEnum.categorical:
            range_type = "list"
        else:
            range_types = [("[1] Min, max", "min_max") , ("[2] List", "list")]
            range_type = inquire_list(f'{message_prefix} - range type', range_types)
        
        # Parameter values
        if range_type == "min_max":
            parameter["range"]["min"] = inquire_int(f'{message_prefix} - minimum')
            parameter["range"]["max"] = inquire_int(f'{message_prefix} - maximum')

            validate_int(parameter["range"]["max"], "Invalid parameter range values.", gt=parameter["range"]["min"])

            if ctx.store['algorithm'] == SweepAlgorithmEnum.grid and \
                    parameter["type"] != SweepSearchSpaceParameterTypeEnum.categorical:
                parameter["step"] = inquire_int(f'{message_prefix} - step')
        else:
            examples = "[64, 128, 256]"
            if parameter["type"] == SweepSearchSpaceParameterTypeEnum.categorical:
                examples = '["sgd", "adam"]'

            range_list_string = inquire_text(f'{message_prefix} - list (ex. {examples})')
            parameter["range"]["list"] = ast.literal_eval(range_list_string)

            if not isinstance(parameter["range"]["list"], list) or len(parameter["range"]["list"]) == 0:
                raise ExitException("Invalid parameter range values.")

            if parameter["type"] == SweepSearchSpaceParameterTypeEnum.int:
                for v in parameter["range"]["list"]:
                    validate_int(v, "Invalid parameter range values.")
            elif parameter["type"] in [SweepSearchSpaceParameterTypeEnum.float, SweepSearchSpaceParameterTypeEnum.double]:
                for v in parameter["range"]["list"]:
                    validate_float(v, "Invalid parameter range values.")

        # Add parameter to list
        search_space_parameters.append(parameter)
        
        # Add another or finish
        if inquire_list("Choose", [
            ("[1] Add another parameter", False),
            ("[2] Finish adding parameters", True),
        ]):
            ctx.store["search_space"] = search_space_parameters
            return


def max_experiment_count_callback(ctx: Context, max_experiment_count: int) -> int:
    ctx.store['max_experiment_count'] = max_experiment_count
    if ctx.store['max_experiment_count'] is None:
        ctx.store['max_experiment_count'] = inquire_int("Maximum number of experiments")

    validate_int(
        ctx.store['max_experiment_count'], 
        f'Invalid max experiment count: {ctx.store["max_experiment_count"]}. Must be greater than 0.',
        gt=0,
    )
    return ctx.store['max_experiment_count']


def parallel_experiment_count_callback(ctx: Context, parallel_experiment_count: int) -> int:
    assert ctx.store['max_experiment_count']

    ctx.store['parallel_experiment_count'] = parallel_experiment_count
    if ctx.store['parallel_experiment_count'] is None:
        ctx.store['parallel_experiment_count'] = inquire_int("Number of experiments to be run in parallel", default=1)

    validate_int(
        ctx.store["parallel_experiment_count"],
        (
            f'Invalid parallel experiment count: {ctx.store["parallel_experiment_count"]}. '
            f'Must be greater than 0, at most {ctx.store["max_experiment_count"]} (max experiment count).'
        ),
        gt=0, lte=ctx.store['max_experiment_count'],
    )
    return ctx.store['parallel_experiment_count']


def max_failed_experiment_count_callback(ctx: Context, max_failed_experiment_count: int) -> int:
    assert ctx.store['max_experiment_count']

    ctx.store['max_failed_experiment_count'] = max_failed_experiment_count
    if ctx.store['max_failed_experiment_count'] is None:
        ctx.store['max_failed_experiment_count'] = inquire_int(
            "Maximum number of experiments to allow to fail",
            default=ctx.store['max_experiment_count']
        )

    validate_int(
        ctx.store["parallel_experiment_count"],
        (
            f'Invalid max failed experiment count: {ctx.store["max_failed_experiment_count"]}. '
            f'Must be greater than 0, at most {ctx.store["max_experiment_count"]} (max experiment count).'
        ),
        gt=0, lte=ctx.store['max_experiment_count'],
    )
    return ctx.store['max_failed_experiment_count']



import os
from typing import Optional

import inquirer
import typer

from savvihub.api.exceptions import NotFoundAPIException
from savvihub.api.uploader import Uploader
from savvihub.api.zipper import Zipper
from savvihub.cli.commands.volume import parse_remote_volume_path
from savvihub.cli.constants import PROJECT_TYPE_CLI_DRIVEN, PROJECT_TYPE_VERSION_CONTROL
from savvihub.cli.exceptions import ExitException, InvalidGitRepository
from savvihub.cli.git import GitRepository
from savvihub.cli.typer import Context


def eval_project_type(project, project_type: str) -> bool:
    if project.type != project_type:
        raise ExitException(f'Project `{project.name}` type `{project.type}` does not match with `{project_type}`')
    return True


def project_name_callback(ctx: Context, project_name: str) -> str:
    organization_name = ctx.params['organization_name']
    try:
        git_repo = GitRepository()
        owner, repo_name, _ = git_repo._get_github_repo()
        if project_name:
            try:
                ctx.project = ctx.authenticated_client.project_read(organization_name, project_name)
            except NotFoundAPIException:
                raise ExitException(f'Project `{project_name}` does not exist in the organization `{organization_name}`.')

            if ctx.project.cached_git_owner_slug == owner and ctx.project.cached_git_repo_slug == repo_name:
                ctx.git_repo = git_repo
        else:
            projects = ctx.authenticated_client.project_list(organization_name).results
            matched_projects = []
            for project in projects:
                if project.cached_git_owner_slug == owner and project.cached_git_repo_slug == repo_name:
                    matched_projects.append(project)

            if len(matched_projects) == 1:
                ctx.project = matched_projects[0]
                ctx.git_repo = git_repo
            elif len(matched_projects) > 1:
                ctx.project = inquirer.prompt([inquirer.List(
                    'project',
                    message='Select project',
                    choices=[(p.name, p) for p in matched_projects],
                )], raise_keyboard_interrupt=True).get('project')
                ctx.git_repo = git_repo
            else:
                ctx.project = inquirer.prompt([inquirer.List(
                    'project',
                    message='Select project',
                    choices=[(p.name, p) for p in projects],
                )], raise_keyboard_interrupt=True).get('project')

        return ctx.project.name

    except InvalidGitRepository:
        if project_name:
            try:
                project = ctx.authenticated_client.project_read(organization_name, project_name)
                if eval_project_type(project, PROJECT_TYPE_CLI_DRIVEN):
                    ctx.project = project
            except NotFoundAPIException:
                raise ExitException(f'Project `{project_name}` does not exist in the organization `{organization_name}`.')
            return project_name
        else:
            projects = ctx.authenticated_client.project_list(organization_name).results
            if not projects:
                raise ExitException(f'No project found in organization `{organization_name}`')

            if len(projects) == 1:
                project = projects[0]
                if eval_project_type(project, PROJECT_TYPE_CLI_DRIVEN):
                    ctx.project = project
            else:
                project = inquirer.prompt([inquirer.List(
                    'project',
                    message='Select project',
                    choices=[(p.name, p) for p in projects],
                )], raise_keyboard_interrupt=True).get('project')
                if eval_project_type(project, PROJECT_TYPE_CLI_DRIVEN):
                    ctx.project = project
            return ctx.project.name


def local_project_callback(ctx: Context, local_project_url: str) -> Optional[str]:
    if ctx.project.type == PROJECT_TYPE_VERSION_CONTROL:
        return None

    if local_project_url:
        volume_id, dataset_version, path = parse_remote_volume_path(local_project_url)
        try:
            file = ctx.authenticated_client.volume_file_read(volume_id, path)
        except NotFoundAPIException:
            raise ExitException(f'Volume file not found: {local_project_url}')
        ctx.store['local_project_uploaded_path'] = file.path
        return file.path
    else:
        # Create zip file in the parent of the current working directory
        cwd = os.path.abspath(os.getcwd())
        zip_path = os.path.abspath(os.path.join(os.path.join(cwd, os.pardir), f'{ctx.project.name}.zip'))
        zipper = Zipper(zip_path, 'w')
        zipper.zipdir(cwd, zipper)
        zipper.close()

        typer.echo('Upload the zipped local project')
        uploaded = Uploader.upload(
            ctx.authenticated_client,
            local_path=zipper.filename,
            volume_id=ctx.project.volume_id,
            remote_path=os.path.basename(zipper.filename),
            progressable=typer.progressbar
        )
        local_project_path = uploaded.path

        os.remove(zipper.filename)

        ctx.store['local_project_uploaded_path'] = local_project_path

        return zipper.filename


project_name_option = typer.Option(None, '--project', callback=project_name_callback,
                                   help='If not present, uses git repository name of the current directory.')



import warnings
import keras
import numpy as np

from savvihub import log, Image
from savvihub.constants import SAVVIHUB_PLOTS_FILETYPE_IMAGE, SAVVIHUB_PLOTS_FILETYPE_IMAGES


class SavviHubCallback(keras.callbacks.Callback):
    def __init__(self, data_type=None, validation_data=None, num_images=None, labels=None, start_epoch=0,
                 save_image=False):
        super().__init__()
        self._data_type = data_type
        self._num_images = num_images or 1
        self._labels = labels
        self._start_epoch = start_epoch
        self._save_image = save_image

        self.validation_data = None
        if validation_data is not None:
            self.validation_data = validation_data

    def _results_to_predicts(self, results):
        predicts = []
        if results[0].shape[-1] == 1:
            if len(self._labels) == 2:
                predicts = [self._lables[1] if result[0] > 0.5 else self._labels[0] for result in results]
            else:
                if not self._labels:
                    warnings.warn('Cannot find labels for prediction')
                predicts = [result[0] for result in results]
        else:
            argmax_results = np.argmax(np.stack(results), axis=1)
            if not self._labels:
                warnings.warn('Cannot find labels for prediction')
                predicts = argmax_results
            else:
                for argmax_result in argmax_results:
                    try:
                        predicts.append(self._labels[argmax_result])
                    except IndexError:
                        predicts.append(argmax_result)
        return predicts

    def _inference(self):
        x_val, y_val = self.validation_data

        if self._num_images > len(x_val):
            self._num_images = len(x_val)

        random_indices = np.random.choice(len(x_val), self._num_images, replace=False)
        x_val_random = [x_val[i] for i in random_indices]
        y_val_random = [y_val[i] for i in random_indices]

        results = self.model.predict(np.stack(x_val_random), batch_size=1)
        predicts = self._results_to_predicts(results)

        captions = []
        for predict, truth in zip(predicts, y_val_random):
            captions.append(f'Pred: {predict} Truth: {truth}')

        return [Image(x, caption=caption) for x, caption in zip(x_val_random, captions)]

    def on_epoch_end(self, epoch, logs=None):
        log(step=epoch+self._start_epoch+1, row=logs)

        if self._save_image and \
                self._data_type in (SAVVIHUB_PLOTS_FILETYPE_IMAGE, SAVVIHUB_PLOTS_FILETYPE_IMAGES):
            if self.validation_data is None:
                warnings.warn('Cannot find validation_data')

            log({
                "validation_image": self._inference(),
            })



from typing import Optional

import inquirer

from savvihub.cli.exceptions import ExitException
from savvihub.cli.typer import Context
from savvihub.cli.utils import inquire_list


def resource_name_callback(ctx: Context, resource_name: str) -> Optional[str]:
    assert ctx.params['organization_name'] and ctx.store['cluster']

    if ctx.store['cluster'].is_savvihub_managed:
        resources = ctx.authenticated_client.kernel_resource_list(ctx.params['organization_name']).results
        if resource_name:
            for resource in resources:
                if resource.name == resource_name.strip():
                    ctx.store['processor_type'] = resource.processor_type
                    ctx.store['resource_spec_id'] = resource.id
                    return resource.name
            else:
                raise ExitException(f'Resource not found: {resource_name.strip()}')
        else:
            selected_resource = inquire_list(
                "Please choose a resource",
                [(f'{x.name} ({x.description})', x) for x in resources],
            )
            ctx.store['processor_type'] = selected_resource.processor_type
            ctx.store['resource_spec_id'] = selected_resource.id
            return selected_resource.name

    elif resource_name:
        raise ExitException('--resource option can be set only with a savvihub-managed cluster')

    return None


def processor_type_callback(ctx: Context, processor_type: str) -> str:
    assert ctx.store['cluster']

    if ctx.store['cluster'].is_savvihub_managed:
        return ctx.store['processor_type']

    processor_type = processor_type.upper() if processor_type else None
    if processor_type not in ['CPU', 'GPU']:
        processor_type = inquire_list(
            "Please choose a processor type",
            ['CPU', 'GPU'],
        )
    return processor_type


def cpu_limit_callback(ctx: Context, cpu_limit: float) -> Optional[float]:
    assert ctx.store['cluster']
    if ctx.store['cluster'].is_savvihub_managed:
        return 0

    if cpu_limit <= 0:
        cpu_limit = float(inquirer.prompt([inquirer.Text(
            'question',
            message="CPU limit (the number of vCPUs)",
            default=1.0,
        )], raise_keyboard_interrupt=True).get('question'))
        if cpu_limit <= 0:
            raise ExitException('Must be greater than 0')

    return cpu_limit


def memory_limit_callback(ctx: Context, memory_limit: str) -> Optional[str]:
    assert ctx.store['cluster']
    if ctx.store['cluster'].is_savvihub_managed:
        return None
    if memory_limit:
        return memory_limit

    memory_limit = float(inquirer.prompt([inquirer.Text(
        'question',
        message="Memory limit in GiB",
        default='4.0',
    )], raise_keyboard_interrupt=True).get('question'))
    if memory_limit <= 0:
        raise ExitException('Must be greater than 0')

    return str(memory_limit) + 'Gi'


def gpu_type_callback(ctx: Context, gpu_type: str) -> str:
    assert ctx.store['cluster'] and ctx.params['processor_type']
    if ctx.store['cluster'].is_savvihub_managed or ctx.params['processor_type'] != 'GPU':
        return 'Empty'

    if gpu_type:
        return gpu_type

    nodes = ctx.authenticated_client.cluster_node_list(
        ctx.params['organization_name'], ctx.store['cluster'].name).nodes
    ctx.store['nodes'] = nodes
    gpu_type = inquire_list(
        "Please choose the GPU type",
        [(f'{x.gpu_product_name}: ({x.gpu_allocatable - x.gpu_limits}/{x.gpu_allocatable})', x) for x in nodes],
    ).gpu_product_name
    return gpu_type


def gpu_limit_callback(ctx: Context, gpu_limit: int) -> int:
    assert ctx.store['cluster'] and ctx.params['gpu_type']
    if ctx.store['cluster'].is_savvihub_managed or ctx.params['gpu_type'] == 'Empty':
        return 0

    if gpu_limit <= 0:
        gpu_limit = int(inquirer.prompt([inquirer.Text(
            'question',
            message="GPU limit (the number of GPUs)",
            default=1,
        )], raise_keyboard_interrupt=True).get('question'))
        if gpu_limit <= 0:
            raise ExitException('Must be greater than 0')
    return gpu_limit



from enum import Enum
from typing import Type

import inquirer
import typer

from openapi_client import ResponseOrganization
from savvihub.api.savvihub import SavviHubClient
from savvihub.api.exceptions import InvalidParametersAPIException, DuplicateAPIException
from savvihub.cli.exceptions import ExitException
from savvihub.cli.constants import DATASET_VERSION_HASH_LATEST


def get_default_organization(client: SavviHubClient) -> ResponseOrganization:
    organizations = client.organization_list().organizations
    if len(organizations) == 0:
        region_list_resp = client.region_list()
        typer.echo('Create organization')
        while True:
            organization_name = inquire_text("Organization name")
            region = inquire_list(
                "Select region",
                [(region.name, region.value) for region in region_list_resp.regions],
                default=region_list_resp.default_region,
            )
            
            try:
                default_organization = client.organization_create(organization_name=organization_name, region=region)
                break
            except InvalidParametersAPIException:
                typer.echo('Invalid organization name. Please try again.')
            except DuplicateAPIException:
                typer.echo('Duplicate organization name exist. Please try again.')

    elif len(organizations) == 1:
        default_organization = organizations[0]
        typer.echo(f'Default organization is automatically set to `{default_organization.name}`.')
        
    else:
        default_organization = inquire_list(
            'Select default organization',
            [(ws.name, ws) for ws in organizations],
        )
        typer.echo(f'Default organization is set to `{default_organization.name}`.')

    return default_organization


def inquire_text(message, default=None):
    key = 'question'
    inquiry = inquirer.Text(key, message=message, default=default)
    return inquirer.prompt([inquiry], raise_keyboard_interrupt=True).get(key)


def inquire_int(message, default=None):
    text = inquire_text(message, default)
    try:
        value = int(text)
    except ValueError:
        raise ExitException(f"Invalid '{text}'. Please enter a valid number.")
    return value


def inquire_float(message, default=None):
    text = inquire_text(message, default)
    try:
        value = float(text)
    except ValueError:
        raise ExitException(f"Invalid '{text}'. Please enter a valid number.")
    return value


def inquire_list(message, choices, default=None):
    key = "question"
    inquiry = inquirer.List(key, message=message, default=default, choices=choices)
    return inquirer.prompt([inquiry], raise_keyboard_interrupt=True).get(key)


def validate_text(value, error_message, nonempty=True, max_length=None):
    if (
        not isinstance(value, str)
        or (nonempty and not value.strip())
        or (max_length is not None and len(value.strip()) > max_length)
    ):
        raise ExitException(error_message)


def validate_int(value, error_message, gt=None, gte=None, lte=None, lt=None):
    # Must satisfy: gt < value < lt, gte <= value <= lte
    if (
        not isinstance(value, int)
        or (gt is not None and gt >= value)
        or (gte is not None and gte > value)
        or (lte is not None and lte < value)
        or (lt is not None and lt <= value)
    ):
        raise ExitException(error_message)


def validate_float(value, error_message, gt=None, gte=None, lte=None, lt=None):
    # Must satisfy: gt < value < lt, gte <= value <= lte
    if (
        not (isinstance(value, float) or isinstance(value, int))
        or (gt is not None and gt >= value)
        or (gte is not None and gte > value)
        or (lte is not None and lte < value)
        or (lt is not None and lt <= value)
    ):
        raise ExitException(error_message)


def validate_enum(value, enum_class: Type[Enum], error_message=None):
    try:
        enum_class(value)
    except ValueError as e:
        raise ExitException(error_message or str(e))


def parse_dataset(dataset_full_name):
    if '@' in dataset_full_name:
        dataset_name, dataset_version_hash = dataset_full_name.split('@', 1)
    else:
        dataset_name = dataset_full_name
        dataset_version_hash = DATASET_VERSION_HASH_LATEST

    if '/' in dataset_name:
        organization_name, dataset_name = dataset_name.split('/', 1)
    else:
        organization_name = None

    return organization_name, dataset_name, dataset_version_hash



from savvihub.cli.exceptions import ExitException
from savvihub.cli.typer import Context
from savvihub.cli.utils import inquire_list, inquire_text, parse_dataset


def dataset_full_name_callback(ctx: Context, dataset_full_name: str) -> str:
    if dataset_full_name:
        _, dataset_name, _ = parse_dataset(dataset_full_name)
        ctx.store['dataset_name'] = dataset_name
        return dataset_full_name

    datasets = ctx.authenticated_client.dataset_list(ctx.organization.name).results
    if len(datasets) == 0:
        raise ExitException(f'Dataset not found in the default organization `{ctx.organization.name}`.\n'
                            f'Specify the dataset name in {{organization}}/{{dataset}} format to select a dataset of another organization.')

    dataset = inquire_list(
        'Select dataset',
        [(d.name, d) for d in datasets],
    )
    ctx.store['dataset_name'] = dataset.name
    return f'{ctx.organization.name}/{dataset.name}'


def download_dest_path_callback(ctx: Context, dest_path: str) -> str:
    if dest_path:
        return dest_path

    dataset_name = ctx.store.get('dataset_name')
    return inquire_text('Dest path', f'./{dataset_name}' if dataset_name else '.')



import subprocess

import inquirer
import typer
from terminaltables import AsciiTable

from savvihub.cli.exceptions import ExitException
from savvihub.cli.inputs.organization import organization_name_option
from savvihub.cli.typer import Typer, Context

project_app = Typer()


@project_app.callback()
def main():
    """
    View projects and pull source codes
    """


@project_app.command(user_required=True)
def list(
    ctx: Context,
    organization_name: str = organization_name_option,
):
    """
    Display a list of projects
    """
    projects_response = ctx.authenticated_client.project_list(organization_name)
    table = AsciiTable([
        ['NAME', 'TYPE', 'REPO'],
        *[[p.name, p.type, p.cached_git_http_url_to_repo or 'N/A']
          for p in projects_response.results],
    ])
    table.inner_column_border = False
    table.inner_heading_row_border = False
    table.inner_footing_row_border = False
    table.outer_border = False

    typer.echo(table.table)


@project_app.command(user_required=True)
def pull(
    ctx: Context,
    organization_name: str = organization_name_option,
):
    """
    Download the project source code
    """
    projects = ctx.authenticated_client.project_list(organization_name).results
    if not projects:
        raise ExitException(f'No version-control project found in organization `{organization_name}`')

    projects = [p for p in projects if p.type == 'version-control']

    if len(projects) == 1:
        project_name = projects[0].name
    else:
        project = inquirer.prompt([inquirer.List(
            'project',
            message='Select project',
            choices=[(p.name, p) for p in projects],
        )], raise_keyboard_interrupt=True).get('project')
        project_name = project.name

    project = ctx.authenticated_client.project_read(organization_name, project_name)
    github_token_response = ctx.authenticated_client.github_token()

    dirname = project.cached_git_repo_slug
    try:
        subprocess.check_output(['git', 'clone', f'https://{github_token_response.token}@github.com/{project.cached_git_owner_slug}/{project.cached_git_repo_slug}.git'])
    except subprocess.CalledProcessError:
        dirname = f'savvihub-{project.cached_git_repo_slug}'
        typer.echo(f'Falling back to \'{dirname}\'...')
        subprocess.check_output(['git', 'clone', f'https://{github_token_response.token}@github.com/{project.cached_git_owner_slug}/{project.cached_git_repo_slug}.git', dirname])

    subprocess.check_output(['rm', '-rf', f'{dirname}/.git'])
